﻿using HarmonyLib;
using HMLLibrary;
using Steamworks;
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Reflection.Emit;
using UnityEngine;
using UnityEngine.UI;
using UnityEngine.SceneManagement;
using Debug = UnityEngine.Debug;
using Object = UnityEngine.Object;
using Random = System.Random;


namespace FullRecycling
{
    public class Main : Mod
    {
        Harmony harmony;
        public static RecycleMode mode = default;
        public static bool VerticalButtons = false;
        public static Random rand = new Random();
        public static Main instance;
        static bool started;
        static bool CanUnload
        {
            get => !instance.modlistEntry.jsonmodinfo.isModPermanent;
            set
            {
                if (instance.modlistEntry.jsonmodinfo.isModPermanent == value)
                {
                    instance.modlistEntry.jsonmodinfo.isModPermanent = !value;
                    ModManagerPage.RefreshModState(instance.modlistEntry);
                }
            }
        }
        public Main() : base()
        {
            instance = this;
        }
        public void Start()
        {
            started = false;
            var unloadButton = modlistEntry.modinfo.unloadBtn.GetComponent<Button>();
            if (SceneManager.GetActiveScene().name != Raft_Network.MenuSceneName)
            {
                unloadButton.onClick.Invoke();
                throw new ModLoadException("Mod must be loaded on the main menu");
            }
            started = true;
            (harmony = new Harmony("com.aidanamite.FullRecycling")).PatchAll();
            Log("Mod has been loaded!");
        }

        public void OnModUnload()
        {
            if (!started)
                return;
            harmony.UnpatchAll(harmony.Id);
            Log("Mod has been unloaded!");
        }

        public override void WorldEvent_WorldLoaded()
        {
            CanUnload = false;
        }

        public override void WorldEvent_WorldUnloaded()
        {
            CanUnload = true;
        }

        public static void LogTree(Transform transform)
        {
            Debug.Log(GetLogTree(transform));
        }

        public static string GetLogTree(Transform transform, string prefix = " -", string step = "--")
        {
            string str = "\n" + prefix + transform.name;
            var c = transform.GetComponents<Component>();
            if (c.Length > 0)
                str += $": {c.Join(x => x.GetType().FullName)}";
            var r = transform.GetComponent<RectTransform>();
            if (r)
                str += $"\n{"".PadLeft(prefix.Length,' ')}rect:[{r.rect.x},{r.rect.y},{r.rect.width},{r.rect.height}]\n{"".PadLeft(prefix.Length, ' ')}anchor:[{r.anchorMin.x},{r.anchorMin.y},{r.anchorMax.x},{r.anchorMax.y}]\n{"".PadLeft(prefix.Length, ' ')}offset:[{r.offsetMin.x},{r.offsetMin.y},{r.offsetMax.x},{r.offsetMax.y}]";
            foreach (Transform sub in transform)
                str += GetLogTree(sub, prefix + step);
            return str;
        }

        public static void UncraftButton_OnClick(SelectedRecipeBox box, PlayerInventory inventory)
        {
            Patch_Slot_RemoveItem.count = 0;
            Patch_Slot_RemoveItem.countUses = true;
            inventory.RemoveItem(box.ItemToCraft.UniqueName, box.ItemToCraft.settings_recipe.AmountToCraft);
            Patch_Slot_RemoveItem.countUses = false;
            var costs = box.ItemToCraft.settings_recipe.NewCost;
            var result = new List<Cost>();
            foreach (var c in costs)
                if (c.items.Length == 1)
                    result.Add(new Cost(c.items[0], c.amount));
                else
                {
                    var col = new List<Cost>();
                    for (int i = 0; i < c.amount; i++)
                    {
                        var item = PickItem(c.items);
                        var r = col.Find(x => x.item.UniqueIndex == item.UniqueIndex);
                        if (r == null)
                            col.Add(new Cost(item, 1));
                        else
                            r.amount++;
                    }
                    result.AddRange(col);
                }
            if (box.ItemToCraft.MaxUses > 1)
            {
                var leave = (int)(result.Sum(x => x.amount) * ((float)Patch_Slot_RemoveItem.count / (box.ItemToCraft.MaxUses * box.ItemToCraft.settings_recipe.AmountToCraft)));
                for (int i = 0; i < result.Count; i++)
                    if (leave <= 0)
                        result[i].amount = 0;
                    else if (leave >= result[i].amount)
                        leave -= result[i].amount;
                    else
                    {
                        result[i].amount -= leave;
                        leave = 0;
                    }
            }
            foreach (var c in result)
                if (c.amount > 0)
                    inventory.AddItem(c.item.UniqueName, c.amount);
        }

        public static Item_Base PickItem(Item_Base[] items)
        {
            if (items.Length == 1 || mode == RecycleMode.First)
                return items[0];
            if (mode == RecycleMode.Last)
                return items[items.Length - 1];
            if (mode == RecycleMode.Random)
                return items[(int)(rand.NextDouble() * items.Length)];
            if (mode == RecycleMode.FirstRandom)
                return items[(int)(Math.Pow(rand.NextDouble(), 2) * items.Length)];
            if (mode == RecycleMode.LastRandom)
                return items[(int)(Math.Sqrt(rand.NextDouble()) * items.Length)];
            return items[0];
        }

        public void ExtraSettingsAPI_Load() => ExtraSettingsAPI_SettingsClose();
        public void ExtraSettingsAPI_SettingsClose()
        {
            mode = (RecycleMode)ExtraSettingsAPI_GetComboboxSelectedIndex("mode");
            VerticalButtons = ExtraSettingsAPI_GetCheckboxState("layout");
        }

        public static int ExtraSettingsAPI_GetComboboxSelectedIndex(string name) => 0;
        public static bool ExtraSettingsAPI_GetCheckboxState(string name) => false;
    }

    public enum RecycleMode
    {
        Random,
        First,
        Last,
        FirstRandom,
        LastRandom
    }

    [HarmonyPatch(typeof(RemovePlaceables), "ReturnItemsFromBlock")]
    static class Patch_ReturnItemsFromBlock
    {
        static IEnumerable<CodeInstruction> Transpiler(IEnumerable<CodeInstruction> instructions)
        {
            var code = instructions.ToList();
            var ind = code.FindLastIndex(x => x.opcode == OpCodes.Ldfld && (x.operand as FieldInfo).DeclaringType == typeof(CostMultiple) && (x.operand as FieldInfo).Name == nameof(CostMultiple.amount));
            if (ind >= 0)
                ind = code.FindIndex(ind, x => x.opcode == OpCodes.Ldc_R4);
            if (ind >= 0)
            {
                code[ind].operand = 1f;
            }
            else
                Debug.LogWarning($"[{Main.instance.name}]: Could not patch \"ReturnItemsFromBlock\". Code instructions missing. Block loot will not be affected by {Main.instance.name}");
            return code;
        }
    }

    [HarmonyPatch(typeof(SelectedRecipeBox), "Initialize")]
    static class Patch_SelectedRecipeBox_Initialize
    {
        static Button UncraftButton;
        static void Prefix(SelectedRecipeBox __instance)
        {
            //Main.LogTree(__instance.transform);
            var r = __instance.GetComponent<RectTransform>();
            var c = r.FindChildRecursively("Craft button").GetComponent<RectTransform>();
            var u = Object.Instantiate(c, c.parent, true);
            u.name = "Uncraft button";
            if (Main.VerticalButtons)
            {
                var f = r.FindChildRecursively("ItemCostList").GetComponent<RectTransform>();
                var i = r.FindChildRecursively("requires text").GetComponent<RectTransform>();
                var s = u.rect.height * 1.1f;
                u.localPosition -= new Vector3(0, s, 0);
                r.offsetMin -= new Vector2(0, s);
                if (i)
                    i.sizeDelta += new Vector2(0, s);
                if (f)
                    f.localPosition -= new Vector3(0, s, 0);
            } else
            {
                var d = r.FindChildRecursively("Divider (1)").GetComponent<RectTransform>();
                var f = r.FindChildRecursively("Skin Grid Layout").parent.GetComponent<RectTransform>();
                var i = r.FindChildRecursively("item_label").GetComponent<RectTransform>();
                var s = u.rect.width * 1.1f;
                u.localPosition += new Vector3(s, 0, 0);
                r.offsetMax += new Vector2(s, 0);
                if (d)
                    d.offsetMax += new Vector2(s, 0);
                if (f)
                    f.sizeDelta += new Vector2(s, 0);
                if (i)
                    i.offsetMax += new Vector2(s, 0);
            }
            Object.DestroyImmediate(u.GetComponentInChildren<I2.Loc.Localize>());
            u.GetComponentInChildren<Text>().text = "Uncraft";
            UncraftButton = u.GetComponent<Button>();
            UncraftButton.onClick = new Button.ButtonClickedEvent();
            UncraftButton.onClick.AddListener(delegate {
                Main.UncraftButton_OnClick(__instance, ComponentManager<PlayerInventory>.Value);
            });
            var b = UncraftButton.gameObject.AddComponent<UncraftButtonUpdater>();
            b.button = UncraftButton;
            b.recipeBox = __instance;
            b.inventory = ComponentManager<PlayerInventory>.Value;
        }
    }

    [HarmonyPatch(typeof(Slot),"RemoveItem")]
    static class Patch_Slot_RemoveItem
    {
        public static bool countUses = false;
        public static int count = 0;
        static void Prefix(Slot __instance,int amount)
        {
            if (!countUses)
                return;
            if (__instance.itemInstance != null && __instance.itemInstance.Amount <= amount)
                count += __instance.itemInstance.UsesInStack;
            else if (__instance.itemInstance != null)
                count += __instance.itemInstance.BaseItemMaxUses * amount;
        }
    }

    [HarmonyPatch(typeof(ModManagerPage), "ShowModInfo")]
    class Patch_ShowModInfo
    {
        static void Prefix(ModData md, ref bool? __state)
        {
            if (md.modinfo.mainClass && md.modinfo.mainClass.GetType() == typeof(Main))
            {
                __state = md.jsonmodinfo.isModPermanent;
                md.jsonmodinfo.isModPermanent = false;
            }
                ModManagerPage.modInfoObj.transform.Find("MakePermanent").gameObject.SetActive(true);
        }
        static void Postfix(ModData md, bool? __state)
        {
            if (__state != null)
                md.jsonmodinfo.isModPermanent = __state.Value;
        }
    }

    public class UncraftButtonUpdater : MonoBehaviour
    {
        public Button button;
        public PlayerInventory inventory;
        public SelectedRecipeBox recipeBox;
        void Update()
        {
            button.interactable = recipeBox.ItemToCraft && inventory.GetItemCount(recipeBox.ItemToCraft) >= recipeBox.ItemToCraft.settings_recipe.AmountToCraft;
        }
    }
}

class ModLoadException : Exception
{
    public ModLoadException(string message) : base(message) { }
}